/**
 * This file was taken from RSTM distribution and modified to use
 * SwissTM.
 * 
 */

// Original copyright notice:

///////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2005, 2006
// University of Rochester
// Department of Computer Science
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//    * Redistributions of source code must retain the above copyright notice,
//      this list of conditions and the following disclaimer.
//
//    * Redistributions in binary form must reproduce the above copyright
//      notice, this list of conditions and the following disclaimer in the
//      documentation and/or other materials provided with the distribution.
//
//    * Neither the name of the University of Rochester nor the names of its
//      contributors may be used to endorse or promote products derived from
//      this software without specific prior written permission.
//
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.

#include "RBTree.h"

using namespace swisstm;
using namespace bench;

// basic constructor
RBNode::RBNode()
    : m_color(BLACK), m_val(-1)
{
}

// helper function:  set the child of node_rw from within a transaction
void setChild(const wr_ptr<RBNode>& node_rw, int cID, const sh_ptr<RBNode>& c)
{
    node_rw->set_child(cID, c);

    if (c != NULL)
    {
        // update c's parent pointer
        wr_ptr<RBNode> c_rw(c);
        c_rw->set_parent(node_rw);
        c_rw->set_ID(cID);
    }
}

/* promote(4):

     2            4
    / \          / \
   1   4   =>   2   5
      / \      / \
     3   5    1   3
 */
// helper function:  promote a node
void promote(wr_ptr<RBNode>& node_rw)
{
    wr_ptr<RBNode> parent_rw(node_rw->get_parent());
    wr_ptr<RBNode> grandparent_rw(parent_rw->get_parent());
    int nodeID = node_rw->get_ID();
    int parentID = parent_rw->get_ID();

    setChild(parent_rw, nodeID, node_rw->get_child(1 - nodeID));
    setChild(node_rw, 1 - nodeID, parent_rw);
    setChild(grandparent_rw, parentID, node_rw);
}

// binary search for the node that has v as its value
bool RBTree::lookup(int v, int &aborts) const
{
    bool found = false;

    BEGIN_TRANSACTION_CA;
    found = false;
    // find v
    rd_ptr<RBNode> sentinel_r(sentinel);
    sh_ptr<RBNode> x = sentinel_r->get_child(0);
    rd_ptr<RBNode> x_r;

    while (x != NULL)
    {
        x_r = x;

        if (x_r->get_val() == v)
        {
            found = true;
            break;
        }
        else
        {
            x = x_r->get_child(v < x_r->get_val() ? 0 : 1);
        }
    }

    END_TRANSACTION_CA;

    return found;
}

// insert a node with v as its value if no such node exists in the tree
void RBTree::insert(int v, int &aborts)
{
    BEGIN_TRANSACTION_CA;

    sh_ptr<RBNode> cx;

    bool found = false;         // found means v is in the tree already
    // find insertion point
    rd_ptr<RBNode> sentinel_r(sentinel);
    rd_ptr<RBNode> x_r(sentinel_r);
    int cID = 0;
    rd_ptr<RBNode> cx_r;

    while (x_r->get_child(cID) != NULL)
    {
        cx_r = x_r->get_child(cID);

        if (cx_r->get_val() == v)
        {
            found = true;
            break; // don't add existing key
        }

        cID = v < cx_r->get_val() ? 0 : 1;
        x_r = cx_r;
    }

    // if the element isn't in the tree, add it and balance the tree
    if (!found)
    {
        wr_ptr<RBNode> x_rw(x_r);
        cx = sh_ptr<RBNode>(new RBNode());
        wr_ptr<RBNode> cx_rw(cx);
        cx_rw->set_color(RED);
        cx_rw->set_val(v);
        setChild(x_rw, cID, cx);

        wr_ptr<RBNode> px_rw;
        wr_ptr<RBNode> y_rw;
        // balance the tree
        while (true)
        {
            x_rw = cx_rw->get_parent();

            if ((x_rw->get_parent() == sentinel) ||
                (BLACK == x_rw->get_color()))
                break;

            px_rw = x_rw->get_parent();
            y_rw = px_rw->get_child(1-x_rw->get_ID());

            if ((y_rw != NULL) && (RED == y_rw->get_color()))
            {
                x_rw->set_color(BLACK);
                y_rw->set_color(BLACK);
                px_rw->set_color(RED);
                cx_rw = px_rw;
            }
            else
            {
                if (cx_rw->get_ID() != x_rw->get_ID())
                {
                    promote(cx_rw);
                    wr_ptr<RBNode> t = cx_rw;
                    cx_rw = x_rw;
                    x_rw = t;
                }

                x_rw->set_color(BLACK);
                px_rw->set_color(RED);
                promote(x_rw);
            }
        }

        sentinel_r = sentinel;

        sh_ptr<RBNode> root = sentinel_r->get_child(0);

        if (root != NULL)
        {
            rd_ptr<RBNode> root_r(root);

            if (root_r->get_color() != BLACK)
            {
                wr_ptr<RBNode> root_rw(root_r);
                root_rw->set_color(BLACK);
            }
        }
    }

    END_TRANSACTION_CA;
}

// remove the node with v as its value if it exists in the tree
void RBTree::remove(int v, int &aborts)
{
    BEGIN_TRANSACTION_CA;

    // find v
    rd_ptr<RBNode> sentinel_r(sentinel);
    sh_ptr<RBNode> x = sentinel_r->get_child(0);
    rd_ptr<RBNode> x_r;

    while (x != NULL)
    {
        x_r = x;

        if (x_r->get_val() == v)
            break;
        else
            x = x_r->get_child(v < x_r->get_val() ? 0 : 1);
    }

    // if we found v, remove it
    if (x != NULL)
    {
        wr_ptr<RBNode> x_rw(x);

        // ensure that we are deleting a node with at most one child
        if ((x_rw->get_child(0) != NULL) &&
            (x_rw->get_child(1) != NULL))
        {
            rd_ptr<RBNode> y_r(x_rw->get_child(1));

            while (y_r->get_child(0)  != NULL)
                y_r = y_r->get_child(0);

            x_rw->set_val(y_r->get_val());
            x = y_r;
            x_rw = x;
        }

        // perform deletion
        sh_ptr<RBNode> p = x_rw->get_parent();
        wr_ptr<RBNode> p_rw(p);
        int cID = (x_rw->get_child(0) != NULL) ? 0 : 1;
        sh_ptr<RBNode> c = x_rw->get_child(cID);
        setChild(p_rw, x_rw->get_ID(), c);

        // fix black height violations
        if ((BLACK == x_rw->get_color()) && (c != NULL))
        {
            rd_ptr<RBNode> c_r(c);

            if (RED == c_r->get_color())
            {
                wr_ptr<RBNode> c_rw(c_r);
                x_rw->set_color(RED);
                c_rw->set_color(BLACK);
            }
        }

        wr_ptr<RBNode> y_rw = x_rw;
        wr_ptr<RBNode> s_rw;

        while ((y_rw->get_parent() != sentinel) &&
               (BLACK == y_rw->get_color()))
        {
            p = y_rw->get_parent();
            p_rw = p;
            // s_rw is a sibling
            s_rw = p_rw->get_child(1 - y_rw->get_ID());

            /* we'd like y's sibling s to be black
             * if it's not, promote it and recolor
             */
            if (RED == s_rw->get_color())
            {
                /*
                    Bp          Bs
                    / \         / \
                  By  Rs  =>  Rp  B2
                      / \     / \
                     B1 B2  By  B1
                 */
                p_rw->set_color(RED);
                s_rw->set_color(BLACK);
                promote(s_rw);
                // reset s
                s_rw = p_rw->get_child(1 - y_rw->get_ID());
            }

            sh_ptr<RBNode> n =
                s_rw->get_child(1 - y_rw->get_ID());
            rd_ptr<RBNode> n_r;

            if (n != NULL)
            {
                n_r = n;

                if (RED == n_r->get_color())
                {
                  farNephew:
                    // goto screws this up... can't assign to n_r because its
                    // not necessarily = n.
                    wr_ptr<RBNode> n_rw(n);
                    /*
                        ?p          ?s
                        / \         / \
                      By  Bs  =>  Bp  Bn
                          / \         / \
                         ?1 Rn      By  ?1
                    */
                    s_rw->set_color(p_rw->get_color());
                    p_rw->set_color(BLACK);
                    n_rw->set_color(BLACK);
                    promote(s_rw);
                    break; // problem solved
                }
                else
                {
                    goto blackn1;
                }
            }
            else
            {
              blackn1:
                n = s_rw->get_child(y_rw->get_ID());

                if (n != NULL)
                {
                    n_r = n;
                    if (RED == n_r->get_color())
                    {
                        wr_ptr<RBNode> n_rw(n_r);
                        /*
                            ?p          ?p
                            / \         / \
                          By  Bs  =>  By  Bn
                              / \           \
                             Rn B1          Rs
                                      \
                                      B1
                        */
                        s_rw->set_color(RED);
                        n_rw->set_color(BLACK);
                        wr_ptr<RBNode> t = s_rw;
                        promote(n_rw);
                        s_rw = n_rw;
                        n = t;
                        goto farNephew; // now the far nephew is red
                    }
                    else
                    {
                        goto blackn2;
                    }
                }
                else
                {
                  blackn2:
                    /*
                        ?p          ?p
                        / \         / \
                      Bx  Bs  =>  Bp  Rs
                          / \         / \
                         B1 B2      B1  B2
                    */
                    s_rw->set_color(RED); // propagate the problem upwards
                }
            }

            // advance to parent and balance again
            y_rw = p_rw;
        }

        // if y was red, this fixes the balance
        y_rw->set_color(BLACK);

        // free storage associated with deleted node
        wr_ptr<RBNode> param1(x_rw->get_parent());
        setChild(param1, x_rw->get_ID(), c);
        // children should NOT be freed
        x_rw->set_child(0, sh_ptr<RBNode>());
        x_rw->set_child(1, sh_ptr<RBNode>());

        tx_delete(x);
    }

    END_TRANSACTION_CA;
}

// returns black-height when balanced and -1 otherwise
static int blackHeight(const sh_ptr<RBNode>& x)
{
    if (x == NULL)
    {
        return 0;
    }
    else
    {
        const rd_ptr<RBNode> x_r(x);
        int bh0 = blackHeight(x_r->get_child(0));
        int bh1 = blackHeight(x_r->get_child(1));
        if ((bh0 >= 0) && (bh1 == bh0))
            return BLACK==x_r->get_color() ? 1+bh0 : bh0;
        else
            return -1;
    }
}

// returns true when a red node has a red child
static bool redViolation(const rd_ptr<RBNode>& p_r, const sh_ptr<RBNode>& x)
{
    if (x == NULL) {
        return false;
    }
    else {
        const rd_ptr<RBNode> x_r(x);
        return ((RED == p_r->get_color() &&
                 RED == x_r->get_color()) ||
                (redViolation(x_r, x_r->get_child(0))) ||
                (redViolation(x_r, x_r->get_child(1))));
    }
}

// returns true when all nodes' parent fields point to their parents
static bool validParents(const sh_ptr<RBNode>& p, int xID,
                         const sh_ptr<RBNode>& x)
{
    if (x == NULL) {
        return true;
    }
    else {
        const rd_ptr<RBNode> x_r(x);
        return ((x_r->get_parent() == p) &&
                (x_r->get_ID() == xID) &&
                (validParents(x, 0, x_r->get_child(0))) &&
                (validParents(x, 1, x_r->get_child(1))));
    }
}

// returns true when the tree is ordered
static bool inOrder(const sh_ptr<RBNode>& x, int lowerBound, int upperBound)
{
    if (x == NULL)
        return true;
    else {
        const rd_ptr<RBNode> x_r(x);
        return ((lowerBound <= x_r->get_val()) &&
                (x_r->get_val() <= upperBound)
                && (inOrder(x_r->get_child(0), lowerBound,
                            x_r->get_val() - 1))
                && (inOrder(x_r->get_child(1),
                            x_r->get_val() + 1, upperBound)));
    }
}

// build an empty tree
RBTree::RBTree()
    : sentinel(new RBNode())
{ }

// sanity check of the RBTree data structure
bool RBTree::isSane() const
{
    bool sane = false;

    BEGIN_TRANSACTION;

    sane = false;
    const rd_ptr<RBNode> sentinel_r(sentinel);
    sh_ptr<RBNode> root = sentinel_r->get_child(0);

    if (root == NULL)
    {
        sane = true; // empty tree needs no checks
    }
    else
    {
        const rd_ptr<RBNode> root_r(root);
        sane = ((BLACK == root_r->get_color()) &&
                (blackHeight(root) >= 0) &&
                !(redViolation(sentinel_r, root)) &&
                (validParents(sentinel, 0, root)) &&
                (inOrder(root, INT_MIN, INT_MAX)));
    }

    END_TRANSACTION;

    return sane;
}

// print a node and recurse on its children
static void printNode(const sh_ptr<RBNode>& x, int indent = 0)
{
    if (x == NULL)
        return;

    const rd_ptr<RBNode> x_r(x);

    printNode(x_r->get_child(0), indent + 2);

    for (int i = 0; i < indent; i++)
        putchar('-');

    printf("%c%d\n",
           RED==x_r->get_color() ? 'R' : 'B',
           x_r->get_val());
    printNode(x_r->get_child(1), indent + 2);

    if (!indent)
        printf("\n\n");
}

// "transaction" that prints a tree; only run this in single-thread mode!
void RBTree::print(int indent) const
{
    BEGIN_TRANSACTION;

    const rd_ptr<RBNode> sentinel_r(sentinel);
    printNode(sentinel_r->get_child(0));

    END_TRANSACTION;
}
