/**
 * All functions related to time are defined in this header file.
 * 
 * @author Aleksandar Dragojevic aleksandar.dragojevic@epfl.ch
 */

#ifndef SWISSTM_TIME_H_
#define SWISSTM_TIME_H_

#include <ctime>
#include <sys/time.h>

#define MILLISECONDS_IN_SECOND 1000
#define MICROSECONDS_IN_SECOND 1000000
#define NANOSECONDS_IN_MILLISECOND 1000000
#define NANOSECONDS_IN_MICROSECOND 1000
#define NANOSECONDS_IN_SECOND 1000000000

namespace swisstm {
	
	/**
	 * Get time in ms since epoch. This function might be architecture
	 * specific.
	 */
	long get_time_ms();

	/**
	 * Get time in us since epoch. This function might be architecture
	 * specific.
	 */
	long long get_time_ns();

	/**
	 * Convenience function for delaying thread for time interval
	 * that is specified in milliseconds.
	 */
	bool sleep(int msec);

	void live_wait_ns(int ns);
}

// define functions as inline here
inline long swisstm::get_time_ms() {
	struct ::timespec t;
	::clock_gettime(CLOCK_REALTIME, &t);
	return t.tv_sec * MILLISECONDS_IN_SECOND +
		t.tv_nsec / NANOSECONDS_IN_MILLISECOND;
}

inline long long swisstm::get_time_ns() {
	struct ::timespec t;
	::clock_gettime(CLOCK_REALTIME, &t);
	return (long long)t.tv_sec * NANOSECONDS_IN_SECOND + t.tv_nsec;
}

inline bool swisstm::sleep(int msec) {
	struct ::timespec req;
	req.tv_sec = msec / MILLISECONDS_IN_SECOND;
	req.tv_nsec = (msec % MILLISECONDS_IN_SECOND)
		* NANOSECONDS_IN_MILLISECOND;

	return ::clock_nanosleep(CLOCK_REALTIME, 0, &req, NULL) != -1;
}

inline void swisstm::live_wait_ns(int wait_interval_ns) {
	long long start_time = get_time_ns();
	long long curr_time;

	do {
		curr_time = get_time_ns();
	} while(curr_time - start_time < wait_interval_ns);
}

#endif
