/**
 * Memory management implementation.
 * 
 * @author Aleksandar Dragojevic aleksandar.dragojevic@epfl.ch
 */

#ifndef SWISSTM_MEMORY_H_
#define SWISSTM_MEMORY_H_

#include <cstdlib>
#include <cstddef>
#include <cassert>
#include <cstring>

#include "../constants.h"
#include "../tid.h"
#include "../common/atomic.h"
#include "../common/tls.h"

#ifdef MM_TBB_MALLOC
#include "scalable_allocator.h"
#endif

#define MAX_MEM_PTR 256

// size of continous part of the pool
#define MEM_PTR_CHUNK_SIZE 1

namespace swisstm {

	class Alloced {
		public:
			virtual ~Alloced() { }
	};

	// These functions are used as a wrapper for real allocate and
	// free functions. This makes it possible to change them easily if
	// needed. For now, they will be just simple malloc and free.
	void *alloc_mem_block(size_t size);
	void free_mem_block(void *mb_ptr);

	// Vector timestamp used for timestamping deleted data.
	struct v_timestamp {
		unsigned long ts[CACHE_LINE_SIZE * MAX_THREADS];
		unsigned len;
	} __attribute__ ((aligned(4096)));

	// Some util functions to use with v_timestamp. These would be a
	// part of timestamp class in a more OO design.
	void copy_v_timestamp(v_timestamp *dest,
		const v_timestamp *src);
	bool is_timestamp_dominated(const v_timestamp *older,
		const v_timestamp *newer);

	// Linked list of memory pointers for deletion.
	struct mem_ptr_arr {
		Alloced *mem_ptr[MAX_MEM_PTR];

		unsigned len;

		v_timestamp vts;

		mem_ptr_arr *next;
	};

	// Some functions that operate on mem_ptr_attr.
	mem_ptr_arr *create_mem_ptr_arr(mem_ptr_arr *next = NULL);
	void free_mem_ptr_arr(mem_ptr_arr *arr);
	bool add_ptr_to_arr(mem_ptr_arr **arr, Alloced *ptr,
		v_timestamp *ts = NULL);
	void free_mem_ptr_arr_list(mem_ptr_arr *arr);

	// Every thread holds a list of objects that are to be deleted.
	struct thread_egc_data {
		mem_ptr_arr *freed;

		// these are temporary
		mem_ptr_arr *free_on_abort;

		mem_ptr_arr *free_on_commit;

		// actual deletes
		mem_ptr_arr *free_on_commit_dominated;

		// don't go through malloc/free all the time - pool arrays instead
		mem_ptr_arr *pool_free_chunk;
		unsigned free_chunk_idx;
		mem_ptr_arr *pool_free_list;

		// TODO: do this in a nicer way
		char pad[64 - 7 * sizeof(void *)];

		static void GlobalInit() {
			// do nothing
		}
	};

	mem_ptr_arr *alloc_pool_free_chunk();

	mem_ptr_arr *alloc_mem_ptr_arr(thread_egc_data *egc);
	void free_mem_ptr_arr(mem_ptr_arr * arr, thread_egc_data *egc);

	// functions that will be invoked by threads/transactions
	void global_init_mem();
	void init_thread_mem();

	// these are non-transactional alloc and free
	void *malloc(size_t size);
	void free(void *);

	thread_egc_data *get_egc_data();

	// Free something on commit. Only elements that are shared should
	// be freed on commit.
	void tx_free_on_commit(Alloced *);

	void tx_free_on_commit_dominated(Alloced *);

	// Free something on abort. Only temporary data, that can become
	// permanent should be freed on abort.	
	void tx_free_on_abort(Alloced *);

	// These are transactional alloc and delete.
	void *tx_malloc(size_t size);
	void *tx_malloc_cloning(size_t size);
	void tx_free(Alloced *);

	// Functions to invoke on transaction start and end.
	void mem_tx_start();
	void mem_tx_commit();
	void mem_tx_abort();

	unsigned get_tx_num(unsigned tid);

	void free_dominated();

	// This is a global timestamp whose components are incremented by
	// different threads.
	extern volatile v_timestamp global_timestamp;
}

// I split declarations and definitions, but want them inline, so they
// are in the header file.

////////////////////////////////
// malloc/free implementation //
////////////////////////////////

inline void *swisstm::alloc_mem_block(size_t size) {

#ifdef MM_TBB_MALLOC
	void *ret = scalable_malloc(size);
#else
	void *ret = ::malloc(size);
#endif

	assert(ret != NULL);

	return ret;
}

inline void swisstm::free_mem_block(void *ptr) {
#ifdef MM_TBB_MALLOC
	scalable_free(ptr);
#else
	::free(ptr);
#endif
}

//////////////////////////
// timestamp operations //
//////////////////////////

inline void swisstm::copy_v_timestamp(swisstm::v_timestamp *dest,
		const swisstm::v_timestamp *src) {
	unsigned len = src->len;
	dest->len = len;

	for(unsigned i = 0;i < len;i++) {
		dest->ts[i * CACHE_LINE_SIZE] = src->ts[i * CACHE_LINE_SIZE];
	}
}

inline bool swisstm::is_timestamp_dominated(const swisstm::v_timestamp *older,
		const swisstm::v_timestamp *newer) {
	unsigned len = older->len;
	bool ret = true;

	for(unsigned i = 0;i < len;i++) {
		unsigned ots = older->ts[i * CACHE_LINE_SIZE];
		unsigned nts = newer->ts[i * CACHE_LINE_SIZE];

		if(!(nts > ots || nts % 2 == 0)) {
			ret = false;
			break;
		}
	}

	return ret;
}

////////////////////////////
// mem_ptr_arr operations //
////////////////////////////

inline swisstm::mem_ptr_arr *swisstm::alloc_pool_free_chunk() {
	return (mem_ptr_arr *)alloc_mem_block(MEM_PTR_CHUNK_SIZE * sizeof(mem_ptr_arr));
}

inline swisstm::mem_ptr_arr *swisstm::alloc_mem_ptr_arr(thread_egc_data *egc) {
	if(egc->free_chunk_idx < MEM_PTR_CHUNK_SIZE) {
		return egc->pool_free_chunk + egc->free_chunk_idx++;
	}

	if(egc->pool_free_list != NULL) {
		mem_ptr_arr *ret = egc->pool_free_list;
		egc->pool_free_list = egc->pool_free_list->next;
		return ret;
	}

	egc->pool_free_chunk = alloc_pool_free_chunk();
	egc->free_chunk_idx = 1;
	return egc->pool_free_chunk;
}

inline void swisstm::free_mem_ptr_arr(mem_ptr_arr * arr, thread_egc_data *egc) {
	arr->next = egc->pool_free_list;
	egc->pool_free_list = arr;
}

inline swisstm::mem_ptr_arr *swisstm::create_mem_ptr_arr(
		swisstm::mem_ptr_arr *next) {
	mem_ptr_arr *ret = alloc_mem_ptr_arr(get_egc_data());
	ret->len = 0;
	ret->next = next;
	return ret;
}

inline void swisstm::free_mem_ptr_arr(swisstm::mem_ptr_arr *arr) {
	free_mem_ptr_arr(arr, get_egc_data());
}

// one function to call when adding pointers to array of pointers
inline bool swisstm::add_ptr_to_arr(swisstm::mem_ptr_arr **arr,
		swisstm::Alloced *ptr, swisstm::v_timestamp *ts) {
	bool ret = false;
	swisstm::mem_ptr_arr *curr = *arr;

	// if mem_ptr is full, create a new one
	if(curr == NULL) {
		curr = swisstm::create_mem_ptr_arr();
		*arr = curr;
	}

	curr->mem_ptr[curr->len] = ptr;
	curr->len++;

	// make sure not to leave empty first mem_ptr_arr
	if(curr->len == MAX_MEM_PTR) {
		if(ts != NULL) {
			copy_v_timestamp(&(curr->vts), ts);
		}

		*arr = swisstm::create_mem_ptr_arr(curr);
		ret = true;
	}

	return ret;
}

inline void swisstm::free_mem_ptr_arr_list(swisstm::mem_ptr_arr *arr) {
	swisstm::mem_ptr_arr *curr = arr;
	swisstm::mem_ptr_arr *prev;

	while(curr != NULL) {
		prev = curr;
		curr = curr->next;
		swisstm::free_mem_ptr_arr(prev);
	}
}

// simple malloc
inline void *swisstm::malloc(size_t size) {
	return swisstm::alloc_mem_block(size);
}

// simple free
inline void swisstm::free(void *ptr) {
	swisstm::free_mem_block(ptr);
}

// get relevant tls data
inline swisstm::thread_egc_data *swisstm::get_egc_data() {
	return Tls<thread_egc_data>::Get();
}

// prepare something to delete if transaction commits
inline void swisstm::tx_free_on_commit(Alloced *ptr) {
	swisstm::thread_egc_data *egc = get_egc_data();
	swisstm::add_ptr_to_arr(&(egc->free_on_commit), ptr);
}

inline void swisstm::tx_free_on_commit_dominated(Alloced *ptr) {
	swisstm::thread_egc_data *egc = get_egc_data();
	swisstm::add_ptr_to_arr(&(egc->free_on_commit_dominated), ptr);
}

// prepate something to delete if transaction aborts
inline void swisstm::tx_free_on_abort(Alloced *ptr) {
	swisstm::thread_egc_data *egc = get_egc_data();
	swisstm::add_ptr_to_arr(&(egc->free_on_abort), ptr);
}

// transactional malloc
inline void *swisstm::tx_malloc(size_t size) {
	void *ptr = swisstm::alloc_mem_block(size);
	swisstm::tx_free_on_abort((Alloced *)ptr);
	return ptr;
}

inline void *swisstm::tx_malloc_cloning(size_t size) {
	void *ptr = swisstm::alloc_mem_block(size);
	swisstm::tx_free_on_abort((Alloced *)ptr);
	return ptr;
}

// transactional free
inline void swisstm::tx_free(Alloced *ptr) {
	swisstm::tx_free_on_commit_dominated(ptr);
}

inline void swisstm::mem_tx_start() {
	global_timestamp.ts[get_tid() * CACHE_LINE_SIZE]++;
}

inline unsigned swisstm::get_tx_num(unsigned tid) {
	return global_timestamp.ts[tid * CACHE_LINE_SIZE];
}

inline void swisstm::mem_tx_commit() {
	// increase global counter
	global_timestamp.ts[get_tid() * CACHE_LINE_SIZE]++;

	// get gc data to perform cleanup
	swisstm::thread_egc_data *egc = swisstm::get_egc_data();

	// first free list of objects to delete on abort
	swisstm::free_mem_ptr_arr_list(egc->free_on_abort);
	egc->free_on_abort = NULL;

	// now put all of the pointers into freed list
	swisstm::mem_ptr_arr *curr = egc->free_on_commit;
	swisstm::mem_ptr_arr *prev;

	while(curr != NULL) {
		unsigned len = curr->len;

		for(unsigned i = 0;i < len;i++) {
			delete(curr->mem_ptr[i]);
		}
		
		prev = curr;
		curr = curr->next;
		swisstm::free_mem_ptr_arr(prev);
	}

	egc->free_on_commit = NULL;

	// actual deletes are the same for both undo and redo logs
	// now put all of the pointers into freed list
	curr = egc->free_on_commit_dominated;

	bool try_to_reclaim = false;

	while(curr != NULL) {
		unsigned len = curr->len;

		for(unsigned i = 0;i < len;i++) {
			if(swisstm::add_ptr_to_arr(&(egc->freed), curr->mem_ptr[i],
					(v_timestamp *)&swisstm::global_timestamp)) {
				try_to_reclaim = true;
			}
		}
		
		prev = curr;
		curr = curr->next;
		swisstm::free_mem_ptr_arr(prev);
	}

	egc->free_on_commit_dominated = NULL;

	// try to actually release some memory
	if(try_to_reclaim) {
		swisstm::free_dominated();
	}
}

// look for something that can be safely released
inline void swisstm::free_dominated() {
	// get gc data to perform cleanup
	swisstm::thread_egc_data *egc = swisstm::get_egc_data();
	swisstm::mem_ptr_arr *curr = egc->freed;
	swisstm::mem_ptr_arr *prev = curr;

	// skip the first as it is never full
	if(curr != NULL) {
		curr = curr->next;
	}

	// now search for the first that can be deleted
	while(curr != NULL) {
		if(swisstm::is_timestamp_dominated(&(curr->vts), 
				(const swisstm::v_timestamp *)&global_timestamp)) {
			break;
		}

		prev = curr;
		curr = curr->next;
	}

	// cut off all dominated arrays, if any
	if(curr != NULL) {
		prev->next = NULL;
	}

	// curr now points to the first one that can be deleted
	while(curr != NULL) {
		unsigned len = curr->len;

		for(unsigned i = 0;i < len;i++) {
			delete(curr->mem_ptr[i]);
		}
		
		prev = curr;
		curr = curr->next;
		swisstm::free_mem_ptr_arr(prev);
	}
}

inline void swisstm::mem_tx_abort() {
	global_timestamp.ts[get_tid() * CACHE_LINE_SIZE]++;

	// get gc data to perform cleanup
	swisstm::thread_egc_data *egc = swisstm::get_egc_data();

	// first free list of objects to delete on commit
	swisstm::free_mem_ptr_arr_list(egc->free_on_commit);
	egc->free_on_commit = NULL;

	// first free list of objects to delete on commit
	swisstm::free_mem_ptr_arr_list(egc->free_on_commit_dominated);
	egc->free_on_commit_dominated = NULL;

	// delete all objects that should be deleted on abort
	swisstm::mem_ptr_arr *curr = egc->free_on_abort;
	swisstm::mem_ptr_arr *prev;

	while(curr != NULL) {
		unsigned len = curr->len;

		for(unsigned i = 0;i < len;i++) {
			delete(curr->mem_ptr[i]);
		}
		
		prev = curr;
		curr = curr->next;
		swisstm::free_mem_ptr_arr(prev);
	}

	egc->free_on_abort = NULL;
}

// Allocator to use with STL containers.
namespace swisstm {
	template <typename T>
	class SwisstmAllocator {
		public:
			typedef T value_type;
			typedef T* pointer;
			typedef const T* const_pointer;
			typedef T& reference;
			typedef const T& const_reference;
			typedef std::size_t size_type;
			typedef std::ptrdiff_t difference_type;

			template <class U>
			struct rebind {
				typedef SwisstmAllocator<U> other;
			};

			pointer address(reference value) const {
				return &value;
			}

			const_pointer address(const_reference value) const {
				return &value;
			}

			SwisstmAllocator() { }

			SwisstmAllocator(const SwisstmAllocator&) { }

			template <class U>
			SwisstmAllocator(const SwisstmAllocator<U>&) { }

			~SwisstmAllocator() { }

			pointer allocate(size_type num, const void* = 0) {
				return static_cast<pointer>(
					swisstm::malloc(num * sizeof(T)));
			}

			void construct(pointer p, const T& value) {
				new((void*)p)T(value);
			}

			void destroy(pointer p) {
				p->~T();
			}

			void deallocate(pointer p, size_type num) {
				swisstm::free(static_cast<void*>(p));
			}
	};

    template <typename T1, typename T2>
    bool operator==(const SwisstmAllocator<T1>&, const SwisstmAllocator<T2>&) {
        return true;
    }

    template <typename T1, typename T2>
    bool operator!=(const SwisstmAllocator<T1>&, const SwisstmAllocator<T2>&) {
        return false;
    }
}

// Convenience class to inherit from when creating transactional
// objects.
namespace swisstm {

	class Cloning {	};

	class SwisstmTxAlloced : public Alloced {
		public:
			void *operator new(size_t size) {
				return tx_malloc(size);
			}

			void *operator new(size_t size, Cloning cloning) {
				return tx_malloc_cloning(size);
			}

			// don't do tx_free, as that will lead to infinite looping
			void operator delete(void* ptr) {
				swisstm::free(ptr);
			}

			virtual ~SwisstmTxAlloced() { }
	};

	class SwisstmAlloced : public Alloced {
		public:
			void* operator new(size_t size) {
				return swisstm::malloc(size);
			}

			void operator delete(void* ptr) {
				swisstm::free(ptr);
			}

			virtual ~SwisstmAlloced() { }
	};
}

#endif // SWISSTM_MEMORY_H_
