/**
 * API layer. This is mostly stripped down RSTM API layer.
 * 
 * @author Aleksandar Dragojevic aleksandar.dragojevic@epfl.ch
 */

#ifndef SWISSTM_TX_PTR_H_
#define SWISSTM_TX_PTR_H_

#include <iostream>

#include "memory.h"
#include "transaction.h"

namespace swisstm {
	template <typename> class sh_ptr;
	template <typename> class rd_ptr;
	template <typename> class wr_ptr;

	template <typename T> void tx_delete(
		sh_ptr<T>& delete_on_tx_commit);
	template <typename T> void tx_delete(
		wr_ptr<T>& delete_on_tx_commit);
	template <typename T> void tx_delete(
		rd_ptr<T>& delete_on_tx_commit);

	template <typename T>
	class sh_ptr {
		public:
			friend class rd_ptr<T>;
			friend class wr_ptr<T>;

			friend void tx_delete<>(sh_ptr<T>&);

		public:
			sh_ptr();
			explicit sh_ptr(T *);
			explicit sh_ptr(ObjectHandle<T> *);
			sh_ptr(const sh_ptr<T>& copy);
			sh_ptr(const wr_ptr<T>& wr);
			sh_ptr(const rd_ptr<T>& rd);

			sh_ptr<T>& operator=(const int must_be_null);
			sh_ptr<T>& operator=(const sh_ptr<T>& assign);

			bool operator==(const void*) const;
			bool operator!=(const void*) const;
			bool operator==(const sh_ptr<T>&) const;
			bool operator!=(const sh_ptr<T>&) const;

			// so that sh_ptr<T>s can be stored in STL collections
			bool operator<(const sh_ptr<T>& rhs) const;

			template <typename T2> operator sh_ptr<T2>();

		private:
			ObjectHandle<T> * volatile handle;
	};

    template <typename T>
    class rd_ptr {
		friend class sh_ptr<T>;
		friend class wr_ptr<T>;

		friend void tx_delete<>(rd_ptr<T>&);

		public:
			rd_ptr(const int must_be_null = 0);
			rd_ptr(const rd_ptr<T>& copy);
			rd_ptr(const wr_ptr<T>& wr);
			explicit rd_ptr(const sh_ptr<T>& to_open);

			rd_ptr<T>& operator=(const int must_be_null);
			rd_ptr<T>& operator=(const rd_ptr<T>& assign);
			rd_ptr<T>& operator=(const sh_ptr<T>& to_open);

			const T* operator->() const;
			const T& operator*() const;

			bool operator==(const void*) const;
			bool operator!=(const void*) const;
			bool operator==(const rd_ptr<T>&) const;
			bool operator!=(const rd_ptr<T>&) const;
			bool operator==(const sh_ptr<T>&) const;
			bool operator!=(const sh_ptr<T>&) const;

		private:
			mutable const T* obj;
    };

    template <typename T>
    class wr_ptr {
		friend class sh_ptr<T>;
		friend class rd_ptr<T>;

		friend void tx_delete<>(wr_ptr<T>&);

		public:
			wr_ptr(const int must_be_null = 0);
			wr_ptr(const wr_ptr<T>&);
			wr_ptr(const rd_ptr<T>&);
			explicit wr_ptr(const sh_ptr<T>&);

			wr_ptr<T>& operator=(const int must_be_null);
			wr_ptr<T>& operator=(const wr_ptr<T>&);
			wr_ptr<T>& operator=(const sh_ptr<T>&);

			T* operator->() const;
			T& operator*() const;

			bool operator==(const void*) const;
			bool operator!=(const void*) const;
			bool operator==(const wr_ptr<T>&) const;
			bool operator!=(const wr_ptr<T>&) const;

		private:
			T* open_shared(const sh_ptr<T> &shared);

		private:
			mutable T* obj;
    };
}

namespace swisstm {

	template <typename T>
	inline sh_ptr<T>::sh_ptr() : handle(NULL) {  }

	template <typename T>
	inline sh_ptr<T>::sh_ptr(ObjectHandle<T> *h) : handle(h) { }

	// this is where I could create a new shared object
	template <typename T>
	inline sh_ptr<T>::sh_ptr(T* o) {
		if(o != NULL) {
			handle = (ObjectHandle<T> *)o->getHandle();
		} else {
			handle = NULL;
		}
	}

	template <typename T>
	inline sh_ptr<T>::sh_ptr(const sh_ptr<T>& ptr)
		: handle(ptr.handle) {	}

	template <typename T>
	inline sh_ptr<T>::sh_ptr(const wr_ptr<T>& wr)
		//: handle(wr.obj->getHandle()) {  }
		: handle(wr.obj == NULL ? NULL :
			((T *)wr.obj)->getHandle()) {  }

	template <typename T>
	inline sh_ptr<T>::sh_ptr(const rd_ptr<T>& rd)
		//: handle(((T *)rd.obj)->getHandle()) {  }
		: handle(rd.obj == NULL ? NULL :
			((T *)rd.obj)->getHandle()) {  }

    template <typename T>
    template <typename T2>
    inline sh_ptr<T>::operator sh_ptr<T2>() {
		return sh_ptr<T2>((ObjectHandle<T2> *)handle);
    }

	template <typename T>
	inline sh_ptr<T>& sh_ptr<T>::operator=(const int must_be_null) {
		assert(must_be_null == 0);
		handle = NULL;
		return *this;
	}

	template <typename T>
	inline sh_ptr<T>& sh_ptr<T>::operator=(const sh_ptr<T>& rhs) {
		handle = rhs.handle;
		return *this;
	}

	template <typename T>
	inline bool sh_ptr<T>::operator==(const void* rhs) const {
		return (handle == rhs);
	}

	template <typename T>
	inline bool sh_ptr<T>::operator!=(const void* rhs) const {
		return (handle != rhs);
	}

	template <typename T>
	inline bool sh_ptr<T>::operator==(const sh_ptr<T>& rhs) const {
		return (handle == rhs.handle);
	}

	template <typename T>
	inline bool sh_ptr<T>::operator!=(const sh_ptr<T>& rhs) const {
		return (handle != rhs.handle);
	}

	template <typename T>
	inline bool sh_ptr<T>::operator<(const sh_ptr<T>& rhs) const {
		return (handle < rhs.handle);
	}

	template <typename T>
	void tx_delete(sh_ptr<T> &ptr) {
		CurrentTransaction::Delete(ptr.handle);
	}
}

namespace swisstm {

	template <typename T>
	inline rd_ptr<T>::rd_ptr(const int must_be_null) : obj(NULL) {
		assert(must_be_null == 0);
	}

	template <typename T>
	inline rd_ptr<T>::rd_ptr(const rd_ptr<T>& ptr)
		: obj(ptr.obj) { }

	template <typename T>
	inline rd_ptr<T>::rd_ptr(const wr_ptr<T>& wr) {
		T *wobj = (T *)wr.obj;

		if(wobj != NULL) {
			obj = (T *)CurrentTransaction::OpenWrite(wobj->getHandle());
		} else {
			obj = NULL;
		}
	}

	template <typename T>
	inline rd_ptr<T>::rd_ptr(const sh_ptr<T>& sh) {
		obj = (const T *)CurrentTransaction::OpenRead(sh.handle);
	}

	template <typename T>
	inline rd_ptr<T>& rd_ptr<T>::operator=(const int must_be_null) {
		assert(must_be_null == 0);
		obj = NULL;
		return *this;
	}

	template <typename T>
	inline rd_ptr<T>& rd_ptr<T>::operator=(const rd_ptr<T>& rhs) {
		obj = rhs.obj;
		return *this;
	}

	template <typename T>
	inline rd_ptr<T>& rd_ptr<T>::operator=(const sh_ptr<T>& rhs) {
		obj = (const T *)CurrentTransaction::OpenRead(rhs.handle);
		return *this;
	}

	template <typename T>
	const T* rd_ptr<T>::operator->() const {
		return obj;
	}

	template <typename T>
	const T& rd_ptr<T>::operator*() const {
		return *obj;
	}

	template <typename T>
	inline bool rd_ptr<T>::operator==(const void* rhs) const {
		return (obj == rhs);
	}

	template <typename T>
	inline bool rd_ptr<T>::operator!=(const void* rhs) const {
		return (obj != rhs);
	}

	template <typename T>
	inline bool rd_ptr<T>::operator==(const rd_ptr<T>& rhs) const {
		return (obj == rhs.obj);
	}

	template <typename T>
	inline bool rd_ptr<T>::operator!=(const rd_ptr<T>& rhs) const {
		return (obj != rhs.obj);
	}

	template <typename T>
	void tx_delete(rd_ptr<T> &ptr) {
		sh_ptr<T> sh(ptr);
		tx_delete(sh);
	}
}

namespace swisstm {

	template <typename T>
	inline wr_ptr<T>::wr_ptr(const int must_be_null) : obj(NULL) {
		assert(must_be_null == 0);
	}

	template <typename T>
	inline wr_ptr<T>::wr_ptr(const wr_ptr<T>& ptr)
		: obj(ptr.obj) { }

	template <typename T>
	inline wr_ptr<T>::wr_ptr(const rd_ptr<T>& rd) {
		T *robj = (T *)rd.obj;

		if(robj != NULL) {
			obj = (T *)CurrentTransaction::OpenWrite(robj->getHandle());
		} else {
			obj = NULL;
		}
	}

	template <typename T>
	inline wr_ptr<T>::wr_ptr(const sh_ptr<T>& ptr) {
		obj = (T *)CurrentTransaction::OpenWrite(ptr.handle);
	}

	template <typename T>
	inline wr_ptr<T>& wr_ptr<T>::operator=(const int must_be_null) {
		assert(must_be_null == 0);
        obj = NULL;
		return *this;
    }

	template <typename T>
	inline wr_ptr<T>& wr_ptr<T>::operator=(const wr_ptr<T>& ptr) {
		obj = ptr.obj;
		return *this;
	}

	template <typename T>
	inline wr_ptr<T>& wr_ptr<T>::operator=(const sh_ptr<T>& ptr) {
		obj = (T *)CurrentTransaction::OpenWrite(ptr.handle);
		return *this;
	}

	template <typename T>
	T* wr_ptr<T>::operator->() const {
		return obj;
	}

	template <typename T>
	T& wr_ptr<T>::operator*() const {
		return *obj;
	}

	template <typename T>
	inline bool wr_ptr<T>::operator==(const void* rhs) const {
		return (obj == rhs);
	}

	template <typename T>
	inline bool wr_ptr<T>::operator!=(const void* rhs) const {
		return (obj != rhs);
	}

	template <typename T>
	inline bool wr_ptr<T>::operator==(const wr_ptr<T>& rhs) const {
		return (obj == rhs.obj);
	}

	template <typename T>
	inline bool wr_ptr<T>::operator!=(const wr_ptr<T>& rhs) const {
		return (obj != rhs.obj);
	}

	template <typename T>
	void tx_delete(wr_ptr<T> &ptr) {
		sh_ptr<T> sh(ptr);
		tx_delete(sh);
	}
}

#endif // SWISSTM_TX_PTR_H_
